import collections as C
import typing as T

from lazyfpl import structures


def team_constraint(lineup: T.Sequence["structures.Player"], n: int) -> bool:
    return max(C.Counter(p.team for p in lineup).values()) <= n


def gkp_def_same_team(
    lineup: T.Sequence["structures.Player"],
) -> bool:
    gkps = {p.team for p in lineup if p.position == "GKP"}
    defs = {p.team for p in lineup if p.position == "DEF"}
    return bool(gkps.intersection(defs))



import os
import typing
import pathlib

import pydantic


class _Env(pydantic.BaseModel):
    backtrace: int = pydantic.Field(
        alias="FPL_BACKTRACE",
        default=3,
        gt=0,
    )
    debug: bool = pydantic.Field(
        alias="FPL_DEBUG",
        default=False,
    )
    lookahead: int = pydantic.Field(
        alias="FPL_LOOKAHEAD",
        default=3,
        gt=0,
    )
    profile: str = pydantic.Field(
        alias="FPL_PROFILE",
        default="",
    )
    teamid: str = pydantic.Field(
        alias="FPL_TEAMID",
        default="4270770",
    )
    db: pathlib.Path = pydantic.Field(
        alias="FPL_DATABASE",
        default=pathlib.Path(".database.sqlite3"),
    )


backtrace: typing.Final = _Env.model_validate(os.environ).backtrace
db: typing.Final = _Env.model_validate(os.environ).db
debug: typing.Final = _Env.model_validate(os.environ).debug
lookahead: typing.Final = _Env.model_validate(os.environ).lookahead
profile: typing.Final = _Env.model_validate(os.environ).profile
teamid: typing.Final = _Env.model_validate(os.environ).teamid



import dataclasses
import datetime
import statistics
import traceback
import typing

import numpy as np
import torch

from lazyfpl import fetch, ml_model, structures


@dataclasses.dataclass(frozen=True)
class PredictionOutcome:
    prediceted: float
    truth: float
    kickoff: datetime.datetime


def backeval(
    player: structures.Player,
    lookahead: int = 1,
    backtrace: int = 3,
    backstep: int = 10,
) -> typing.Iterator[PredictionOutcome]:
    with torch.no_grad():
        net = ml_model.load_model(player)
        net.eval()
        for n in range(backstep):
            fixutres = [f for f in player.fixutres if not f.upcoming][
                -(backtrace + lookahead + n) : -(lookahead + n)
            ]
            inference = [ml_model.features(f) for f in fixutres]
            try:
                next_fixture = player.fixutres[player.fixutres.index(fixutres[-1]) + 1]
            except IndexError:
                break
            inf = np.expand_dims(
                np.stack(
                    [
                        np.array(dataclasses.astuple(x), dtype=np.float32)
                        for x in inference
                    ],
                    axis=0,
                ).astype(np.float32),
                axis=0,
            )
            xP = net(torch.from_numpy(inf)).detach().numpy()[0]
            assert xP is not None
            assert next_fixture.points is not None
            yield PredictionOutcome(
                prediceted=xP,
                truth=next_fixture.points,
                kickoff=next_fixture.kickoff_time,
            )


def players_backeval() -> dict[structures.Player, tuple[PredictionOutcome, ...]]:
    rv = dict[structures.Player, tuple[PredictionOutcome, ...]]()
    for player in sorted(fetch.players(), key=lambda x: (x.team, x.webname, x.name)):
        if player.xP and player.xP > 0 and player.mtm() > 30:
            try:
                rv[player] = tuple(backeval(player))
            except ValueError as e:
                traceback.print_exception(e)
    return rv


if __name__ == "__main__":
    player_xp = players_backeval()

    for player, ev in player_xp.items():
        print(f"\n{player.name}({player.webname}) - {player.position} - {player.team}")
        for e in sorted(ev, key=lambda x: x.kickoff):
            print(
                f"  When: {e.kickoff.date()} xP: {e.prediceted:<6.1f} TP: {e.truth:<6.1f} "
                f"Err: {abs(e.prediceted - e.truth):<6.1f}"
            )

    print()
    rms = (
        statistics.mean(
            (v.prediceted - v.truth) ** 2
            for values in player_xp.values()
            for v in values
        )
        ** 0.5
    )
    print(f"RMS: {rms:.1f}")
    am = statistics.mean(
        (abs(v.prediceted - v.truth)) for values in player_xp.values() for v in values
    )
    print(f"AM : {am:.1f}")
    for n in range(1, 6):
        errs = (
            1 if abs(v.prediceted - v.truth) <= n else 0
            for values in player_xp.values()
            for v in values
        )
        print(f"RC{n}: {statistics.mean(errs)*100:.1f}")

    print()

    def key(values: tuple[PredictionOutcome, ...]) -> float:
        return statistics.mean((v.prediceted - v.truth) ** 2 for v in values) ** 0.5 + 1

    for player, values in sorted(
        player_xp.items(),
        key=lambda x: (x[0].xP or 0) / (key(x[1])),
    )[-10:]:
        if player.xP:
            print(
                f"{player.webname:<20} {player.xP:<6.2f} "
                f"{key(values):<6.2f} {(player.xP or 0)/(key(values)):<6.2f}"
            )



import csv
import datetime
import argparse
import functools
import io
import traceback
import urllib.parse as up

import pydantic
import pytz
import requests
from dateutil.parser import parse as dtparser
from tqdm.std import tqdm

from lazyfpl import conf, database, structures


def now_tz_utc() -> datetime.datetime:
    return datetime.datetime.now(tz=pytz.utc)


@functools.cache
def bootstrap() -> dict:
    return requests.get(
        "https://fantasy.premierleague.com/api/bootstrap-static/"
    ).json()


@functools.cache
def summary(id: int) -> dict:
    assert isinstance(id, int)
    return requests.get(
        f"https://fantasy.premierleague.com/api/element-summary/{id}/"
    ).json()


@functools.cache
def db_name_pid() -> dict[str, int]:
    return {
        row["name"]: row["id"]
        for row in database.execute(
            """
        SELECT
            id,
            name
        FROM
            player
        """
        )
    }


@functools.cache
def player_id_fuzzer(name: str) -> int:
    for dname, pid in db_name_pid().items():
        if dname.casefold() == name.casefold():
            return pid
        if all(n in dname.casefold() for n in name.casefold().split()):
            return pid
        if all(n in name.casefold() for n in dname.casefold().split()):
            return pid

    raise KeyError(name)


def session_from_url(url: str) -> str:
    return up.urlparse(url).path.split("/")[5]


@functools.cache
def past_team_lists() -> dict[str, list["structures.Team"]]:
    urls = (
        "https://raw.githubusercontent.com/vaastav/Fantasy-Premier-League/master/data/2023-24/teams.csv",
        "https://raw.githubusercontent.com/vaastav/Fantasy-Premier-League/master/data/2022-23/teams.csv",
        "https://raw.githubusercontent.com/vaastav/Fantasy-Premier-League/master/data/2021-22/teams.csv",
    )

    return {
        session_from_url(url): [
            structures.Team.model_validate(t)
            for t in csv.DictReader(io.StringIO(requests.get(url).text))
        ]
        for url in urls
    }


@functools.cache
def past_team_lookup(tid: int, session: database.SESSIONS) -> str:
    assert isinstance(tid, int)
    for team in past_team_lists()[session]:
        if team.id == tid:
            return team.name
    raise ValueError(f"No match: {tid} / {session}.")


@functools.cache
def past_game_lists() -> dict[str, list[dict]]:
    urls = (
        "https://raw.githubusercontent.com/vaastav/Fantasy-Premier-League/master/data/2023-24/gws/merged_gw.csv",
        "https://raw.githubusercontent.com/vaastav/Fantasy-Premier-League/master/data/2022-23/gws/merged_gw.csv",
        "https://raw.githubusercontent.com/vaastav/Fantasy-Premier-League/master/data/2021-22/gws/merged_gw.csv",
    )
    return {
        session_from_url(url): list(csv.DictReader(io.StringIO(requests.get(url).text)))
        for url in urls
    }


def initialize_database() -> None:
    database.execute(
        """
        CREATE TABLE team (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            name TEXT NOT NULL,
            web_team_id INTEGER NOT NULL,
            session TEXT NOT NULL,
            strength_attack_away INTEGER NOT NULL,
            strength_attack_home INTEGER NOT NULL,
            strength_defence_away INTEGER NOT NULL,
            strength_defence_home INTEGER NOT NULL,
            strength_overall_away INTEGER NOT NULL,
            strength_overall_home INTEGER NOT NULL,
            UNIQUE(name, session)
        )
    """
    )
    database.execute(
        """
        CREATE TABLE player (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            webname TEXT NOT NULL,
            name TEXT NOT NULL,
            price INTEGER NOT NULL,
            team_id INTEGER NOT NULL,
            news TEXT NOT NULL,
            model BLOB,
            FOREIGN KEY(team_id) REFERENCES team(id),
            UNIQUE(webname, name, team_id)
        );
    """
    )
    database.execute(
        """
        CREATE TABLE game (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            is_home INTEGER NOT NULL,
            kickoff REAL NOT NULL,
            minutes INTEGER,
            opponent INTEGER NOT NULL,
            player_id INTEGER NOT NULL,
            points INTEGER,
            position TEXT NOT NULL,
            session TEXT NOT NULL,
            team INTEGER NOT NULL,
            upcoming INTEGER NOT NULL,
            FOREIGN KEY(opponent) REFERENCES team(id),
            FOREIGN KEY(player_id) REFERENCES player(id),
            FOREIGN KEY(team) REFERENCES team(id)
        );
    """
    )


def nuke_database() -> None:
    database.execute("""DROP TABLE game;""")
    database.execute("""DROP TABLE player;""")
    database.execute("""DROP TABLE team;""")


def populate_teams() -> None:
    sql = """
        INSERT INTO team (
            name,
            session,
            web_team_id,
            strength_attack_away,
            strength_attack_home,
            strength_defence_away,
            strength_defence_home,
            strength_overall_away,
            strength_overall_home
        ) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?);
    """
    for session, teams in past_team_lists().items():
        for team in teams:
            database.execute(
                sql,
                (
                    team.name,
                    session,
                    team.id,
                    team.strength_attack_away,
                    team.strength_attack_home,
                    team.strength_defence_away,
                    team.strength_defence_home,
                    team.strength_overall_away,
                    team.strength_overall_home,
                ),
            )


def populate_players(session: database.SESSIONS = database.CURRENT_SESSION) -> None:
    sql = """
        INSERT INTO player(
            webname,
            name,
            price,
            news,
            team_id
        ) VALUES (
            ?, ?, ?, ?,
            (SELECT id FROM team WHERE session = ? AND web_team_id = ?)
        );
    """
    for ele in tqdm(
        bootstrap()["elements"],
        ascii=True,
        bar_format="{percentage:3.0f}% | {bar:20} {r_bar}",
        unit_divisor=1_000,
        unit_scale=True,
    ):
        database.execute(
            sql,
            (
                ele["web_name"],
                f'{ele["first_name"]} {ele["second_name"]}',
                ele["now_cost"],
                ele["news"],
                session,
                ele["team"],
            ),
        )


def populate_games() -> None:
    game_sql = """
        INSERT INTO game (
            session,
            upcoming,
            is_home,

            kickoff,
            minutes,
            points,

            position,
            player_id,
            team,
            opponent
        ) VALUES (
            ?, ?, ?, ?, ?, ?, ?,
            ?,
            (SELECT id FROM team WHERE session = ? AND name = ?),
            (SELECT id FROM team WHERE session = ? AND name = ?)
        );
    """
    no_pid = set[str]()  # Print once per player name, avoid cli-spam.
    for session, games in past_game_lists().items():
        for game in tqdm(
            games,
            ascii=True,
            bar_format="{percentage:3.0f}% | {bar:20} {r_bar}",
            unit_divisor=1_000,
            unit_scale=True,
        ):
            # Dafuq pydantic?!
            game.pop(None, None)

            try:
                historic = structures.HistoricGame.model_validate(game)
            except pydantic.ValidationError as e:
                if conf.debug:
                    traceback.print_exception(e)
                continue

            try:
                pid = player_id_fuzzer(historic.name)
            except KeyError as e:
                if conf.debug and historic.name not in no_pid:
                    traceback.print_exception(e)
                no_pid.add(historic.name)
                continue

            try:
                database.execute(
                    game_sql,
                    (
                        session,
                        False,
                        historic.was_home,
                        historic.kickoff_time,
                        historic.minutes,
                        historic.total_points,
                        historic.position,
                        pid,
                        session,
                        historic.team,
                        session,
                        past_team_lookup(historic.opponent_team, session),
                    ),
                )
            except database.sqlite3.IntegrityError as e:
                # The player does not play this year.
                if conf.debug:
                    traceback.print_exception(e)
                continue

    database.execute(
        """
        UPDATE
            game
        SET
            position = "GKP"
        WHERE
            position = "GK"
    """
    )

    now = now_tz_utc()
    upcoming_games: list[int] = [
        e["id"] for e in bootstrap()["events"] if dtparser(e["deadline_time"]) > now
    ]
    for ele in tqdm(
        bootstrap()["elements"],
        ascii=True,
        bar_format="{percentage:3.0f}% | {bar:20} {r_bar}",
        unit_divisor=1_000,
        unit_scale=True,
    ):
        fullname = f'{ele["first_name"]} {ele["second_name"]}'
        team = upcoming_team_id_to_name(ele["team"])
        for game in summary(ele["id"])["fixtures"]:
            try:
                upcoming = structures.UpcommingGame.model_validate(game)
            except pydantic.ValidationError as e:
                if conf.debug:
                    traceback.print_exception(e)
                continue

            # A past game, data allready logged.
            if upcoming.event in upcoming_games:
                team_h = upcoming_team_id_to_name(upcoming.team_h)
                team_a = upcoming_team_id_to_name(upcoming.team_a)
                is_home = team == team_h
                opponent = list({team, team_a, team_h} - {team})[0]
                database.execute(
                    game_sql,
                    (
                        database.CURRENT_SESSION,
                        True,
                        is_home,
                        upcoming.kickoff_time,
                        None,
                        None,
                        upcoming_position(fullname),
                        player_id_fuzzer(fullname),
                        database.CURRENT_SESSION,
                        team,
                        database.CURRENT_SESSION,
                        opponent,
                    ),
                )


@functools.cache
def upcoming_team_id_to_name(id: int) -> str:
    for item in bootstrap()["teams"]:
        if id == item["id"]:
            return item["name"]
    raise ValueError(f"No team name: {id}")


@functools.cache
def upcoming_position(name: str) -> database.POSITIONS:
    for element in bootstrap()["elements"]:
        if f'{element["first_name"]} {element["second_name"]}' == name:
            element_type = bootstrap()["element_types"][element["element_type"] - 1]
            return element_type["singular_name_short"]
    raise ValueError(f"No player named: {name}")


def main():
    parser = argparse.ArgumentParser(
        prog="Populator",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("--fresh", action="store_true")
    parsed = parser.parse_args()

    if parsed.fresh:
        nuke_database()

    initialize_database()
    populate_teams()
    populate_players()
    populate_games()


if __name__ == "__main__":
    main()



import dataclasses
import datetime
import itertools
import statistics
import typing as T

import pydantic

from lazyfpl import conf, database, helpers


@dataclasses.dataclass(frozen=True)
class Difficulty:
    attack: float
    defence: float
    overall: float

    @property
    def mean(self):
        return statistics.mean((self.attack, self.defence, self.overall))


@dataclasses.dataclass(frozen=True)
class Fixture:
    at_home: bool
    kickoff_time: datetime.datetime
    minutes: T.Optional[int]
    opponent: str
    player: str
    points: T.Optional[int]
    session: database.SESSIONS
    team: str
    upcoming: bool
    webname: str
    team_strength_attack_home: int
    team_strength_attack_away: int
    team_strength_defence_home: int
    team_strength_defence_away: int
    team_strength_overall_home: int
    team_strength_overall_away: int
    opponent_strength_attack_home: int
    opponent_strength_attack_away: int
    opponent_strength_defence_home: int
    opponent_strength_defence_away: int
    opponent_strength_overall_home: int
    opponent_strength_overall_away: int

    @property
    def relative(self) -> Difficulty:
        if self.at_home:
            attack = (
                self.team_strength_attack_home - self.opponent_strength_defence_away
            ) / (self.team_strength_attack_home + self.opponent_strength_defence_away)
            defence = (
                self.team_strength_defence_home - self.opponent_strength_attack_away
            ) / (self.team_strength_defence_home + self.opponent_strength_attack_away)
            overall = (
                self.team_strength_overall_home - self.opponent_strength_overall_away
            ) / (self.team_strength_overall_home + self.opponent_strength_overall_away)
        else:
            attack = (
                self.team_strength_attack_away - self.opponent_strength_defence_home
            ) / (self.team_strength_attack_away + self.opponent_strength_defence_home)
            defence = (
                self.team_strength_defence_away - self.opponent_strength_attack_home
            ) / (self.team_strength_defence_away + self.opponent_strength_attack_home)
            overall = (
                self.team_strength_overall_away - self.opponent_strength_overall_home
            ) / (self.team_strength_overall_away + self.opponent_strength_overall_home)

        return Difficulty(
            attack=attack,
            defence=defence,
            overall=overall,
        )


@dataclasses.dataclass(eq=True, unsafe_hash=True)
class Player:
    fixutres: list[Fixture] = dataclasses.field(compare=False, repr=False)
    name: str = dataclasses.field(compare=True)
    news: str = dataclasses.field(compare=True)
    position: database.POSITIONS = dataclasses.field(compare=False)
    price: int = dataclasses.field(compare=False)
    team: str = dataclasses.field(compare=False)
    webname: str = dataclasses.field(compare=True)
    xP: T.Optional[float]

    def tp(self, session: database.SESSIONS = database.CURRENT_SESSION) -> int:
        return sum(f.points or 0 for f in self.fixutres if f.session == session)

    def mtm(self, last: int = conf.backtrace) -> float:
        # fixutre.minutes is None if the gameweek is not started
        # need to filter out fixutres with minutes as none.
        try:
            return statistics.mean(
                (
                    f.minutes or 0
                    for f in sorted(
                        (f for f in self.fixutres if f.minutes is not None),
                        key=lambda x: x.kickoff_time,
                    )[-last:]
                )
            )
        except statistics.StatisticsError:
            return 0.0

    def upcoming_difficulty(self) -> float:
        upcoming = [f for f in self.fixutres if f.upcoming][: conf.lookahead]
        return statistics.mean(f.relative.mean for f in upcoming)

    @property
    def next_opponent(self) -> str:
        return min(
            (f for f in self.fixutres if f.upcoming),
            key=lambda f: f.kickoff_time,
        ).opponent

    def upcoming_opponents(self) -> list[str]:
        return [
            x.opponent
            for x in sorted(
                (f for f in self.fixutres if f.upcoming),
                key=lambda f: f.kickoff_time,
            )
        ]

    def __str__(self):
        return (
            f"{(self.xP or 0):<6.1f} {self.price:<6.1f} {self.tp():<4} "
            f"{self.upcoming_difficulty()*10:<8.1f} {self.team:<15} "
            f"{self.position:<9} {self.webname:<20} "
            f"{' - '.join(self.upcoming_opponents()[:conf.lookahead])} "
            f"{self.news}"
        )


class Squad:
    def __init__(self, players: T.Sequence[Player]):
        self.players = players

    def price(self) -> int:
        return helpers.squad_price(self.players)

    def best_lineup(self) -> list[Player]:
        return helpers.best_lineup(self.players)

    def SxP(self) -> float:
        # Squad xP
        return helpers.squad_xP(self.players)

    def LxP(self) -> float:
        # Lineup xP
        return helpers.squad_xP(self.best_lineup())

    def CxP(self) -> float:
        # Combined xP
        return (self.SxP() ** 2 + self.LxP() ** 2) ** 0.5

    def tscore(self) -> int:
        return helpers.tcnt(self.players)

    def sscore(self, n: int = conf.lookahead) -> int:
        # "sscore -> "schedule score"
        return helpers.sscore(self.players, n=n)

    def tsscore(self, n: int = conf.lookahead) -> float:
        return helpers.tsscore(self.players)

    def __iter__(self):
        yield from self.players

    def __len__(self) -> int:
        return len(self.players)

    def __str(self):
        pospri = {"GKP": 0, "DEF": 1, "MID": 2, "FWD": 3}
        yield f"Price: {self.price()/10} Size: {len(self.players)}"
        yield f"LxP: {self.LxP():.1f} SxP: {self.SxP():.1f} CxP: {self.CxP():.1f}"
        yield (
            f"Schedule score: {self.sscore()} Team score: {self.tscore()} "
            f"TSscore: {self.tsscore():.2f}"
        )
        yield (
            "BIS  xP     Price  TP   UD       Team            Position  Player"
            + " " * 15
            + "Upcoming"
            + " " * 30
            + "News"
        )
        for _, players in itertools.groupby(
            sorted(
                self.players,
                key=lambda x: (pospri.get(x.position), x.xP or 0),
                reverse=True,
            ),
            key=lambda x: x.position,
        ):
            for player in players:
                yield ("X    " if player in self.best_lineup() else "     ") + str(
                    player
                )

    def __str__(self) -> str:
        return "\n".join(self.__str())


class HistoricGame(pydantic.BaseModel):
    name: str
    position: database.POSITIONS
    team: str
    xP: float
    assists: int
    bonus: int
    bps: int
    clean_sheets: int
    creativity: float
    element: int
    fixture: float
    goals_conceded: float
    goals_scored: float
    ict_index: float
    influence: float
    kickoff_time: datetime.datetime
    minutes: int
    opponent_team: int
    own_goals: int
    penalties_missed: int
    penalties_saved: int
    red_cards: int
    round: int
    saves: int
    selected: int
    team_a_score: int
    team_h_score: int
    threat: float
    total_points: int
    transfers_balance: int
    transfers_in: int
    transfers_out: int
    value: int
    was_home: bool
    yellow_cards: int
    GW: int

    @pydantic.validator("position", pre=True)
    def gk_to_gkp(cls, value: str) -> str:
        return "GKP" if value == "GK" else value


class UpcommingGame(pydantic.BaseModel):
    code: int
    difficulty: int
    event_name: str | None
    event: int | None
    finished: bool
    id: int
    is_home: bool
    kickoff_time: datetime.datetime | None
    minutes: int
    provisional_start_time: bool
    team_a_score: int | None
    team_a: int
    team_h_score: int | None
    team_h: int


class Team(pydantic.BaseModel):
    code: int
    draw: int
    id: int
    loss: int
    name: str
    played: int
    points: int
    position: int
    pulse_id: int
    short_name: str
    strength: int
    strength_attack_away: int
    strength_attack_home: int
    strength_defence_away: int
    strength_defence_home: int
    strength_overall_away: int
    strength_overall_home: int
    unavailable: bool
    win: int



import argparse
import itertools

from lazyfpl import fetch, structures

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--top",
        "-t",
        default=None,
        help="Top N players per position. (default: %(default)s)",
        type=int,
    )
    args = parser.parse_args()

    players = list(
        itertools.chain(
            *[
                list(p)[: args.top] if args.top else list(p)
                for _, p in itertools.groupby(
                    sorted(
                        fetch.players(),
                        key=lambda x: (x.position, x.xP or 0),
                        reverse=True,
                    ),
                    key=lambda x: x.position,
                )
            ]
        )
    )

    print(structures.Squad(players))



import dataclasses
import datetime
import functools
import itertools
import traceback

import dateutil.parser
import requests

from lazyfpl import conf, database, ml_model, structures


@dataclasses.dataclass(frozen=True)
class Persona:
    first: str
    second: str
    webname: str

    @property
    def combined(self) -> str:
        return " ".join((self.first, self.second))


@functools.cache
def bootstrap() -> dict:
    return requests.get(
        "https://fantasy.premierleague.com/api/bootstrap-static/"
    ).json()


def bootstrap_events() -> list[dict]:
    return bootstrap()["events"]


def next_gw() -> int:
    for e in bootstrap_events():
        if e["is_next"]:
            return int(e["id"])
    raise ValueError


def current_gw() -> int:
    for e in bootstrap_events():
        if e["is_current"]:
            return int(e["id"])
    raise ValueError


def next_deadline() -> datetime.timedelta:
    for e in bootstrap_events():
        if e["is_next"]:
            deadline = dateutil.parser.parse(e["deadline_time"])
            now = datetime.datetime.now(tz=datetime.timezone.utc)
            return deadline - now
    raise ValueError


@functools.cache
def person(pid: int) -> Persona:
    for element in bootstrap()["elements"]:
        if element["id"] == pid:
            return Persona(
                first=element["first_name"],
                second=element["second_name"],
                webname=element["web_name"],
            )
    raise ValueError(f"No player: {pid}")


@functools.cache
def players() -> list["structures.Player"]:
    pool = list[structures.Player]()

    for (name, webname), _games in itertools.groupby(
        sorted(database.games(), key=lambda x: (x.player, x.webname)),
        lambda x: (x.player, x.webname),
    ):
        games = sorted(_games, key=lambda x: x.kickoff)
        fixtures = [
            structures.Fixture(
                at_home=game.is_home,
                kickoff_time=game.kickoff,
                minutes=game.minutes,
                opponent=game.opponent,
                player=name,
                points=game.points,
                session=game.session,
                team=game.team,
                upcoming=game.upcoming,
                webname=webname,
                team_strength_attack_home=game.team_strength_attack_home,
                team_strength_attack_away=game.team_strength_attack_away,
                team_strength_defence_home=game.team_strength_defence_home,
                team_strength_defence_away=game.team_strength_defence_away,
                team_strength_overall_home=game.team_strength_overall_home,
                team_strength_overall_away=game.team_strength_overall_away,
                opponent_strength_attack_home=game.opponent_strength_attack_home,
                opponent_strength_attack_away=game.opponent_strength_attack_away,
                opponent_strength_defence_home=game.opponent_strength_defence_home,
                opponent_strength_defence_away=game.opponent_strength_defence_away,
                opponent_strength_overall_home=game.opponent_strength_overall_home,
                opponent_strength_overall_away=game.opponent_trength_overall_away,
            )
            for game in games
        ]

        try:
            team = [g for g in games if g.upcoming][-1].team
        except IndexError as e:
            if conf.debug:
                traceback.print_exception(e)
            continue
        pool.append(
            structures.Player(
                fixutres=fixtures,
                name=name,
                news=games[-1].news,
                position=games[-1].position,
                price=database.price(games[-1].player_id),
                team=team,
                webname=database.webname(games[-1].player_id),
                xP=None,
            )
        )

    for p in pool:
        try:
            p.xP = ml_model.xP(p)
        except ValueError as e:
            if conf.debug:
                traceback.print_exception(e)

    return pool


@functools.cache
def picks() -> list[dict]:
    if not conf.teamid or not conf.profile:
        raise RuntimeError(
            "Env. FPL-teamid and FPL-cookie/profile must be set. FPL-team id "
            "from URL and cookie 'pl_profile' from 'application' in chrome."
        )

    return requests.get(
        f"https://fantasy.premierleague.com/api/my-team/{conf.teamid}/",
        cookies={"pl_profile": conf.profile},
    ).json()["picks"]


def my_team() -> structures.Squad:
    webnames = {person(pick["element"]).webname for pick in picks()}
    return structures.Squad([p for p in players() if p.webname in webnames])


if __name__ == "__main__":
    import sys

    to_show = sys.argv[1:]
    for p in sorted(players(), key=lambda x: (x.team, x.webname)):
        if p.webname in to_show or p.team in to_show:
            print(f"{p.webname} ({p.team})")
            for f in sorted(
                (f for f in p.fixutres if f.points is not None),
                key=lambda x: x.kickoff_time,
                reverse=True,
            )[: conf.backtrace]:
                vs = (
                    f"{f.opponent} vs. {p.team}"
                    if f.at_home
                    else f"{p.team} vs. {f.opponent}"
                )
                print(
                    f"  {vs:<30} Points: {f.points} "
                    "Minutes: {f.minutes} Diff: {f.relative.mean:.1f}"
                )



import argparse
import heapq
import itertools
import typing as T

from tqdm.std import tqdm
from lazyfpl import constraints, fetch, helpers, structures


def display(
    old: T.Sequence["structures.Player"],
    new: T.Sequence["structures.Player"],
    log: tqdm,
) -> None:
    transfers_in = sorted((p for p in new if p not in old), key=lambda x: x.position)
    transfers_out = sorted((p for p in old if p not in new), key=lambda x: x.position)

    max_len_in_name = max(len(p.webname) for p in transfers_in)
    max_len_in_team = max(len(p.team) for p in transfers_in)

    max_len_out_name = max(len(p.webname) for p in transfers_out)
    max_len_out_team = max(len(p.team) for p in transfers_out)

    log.write("-" * 75)
    for o, i in zip(transfers_out, transfers_in):
        log.write(
            f"{o.position}: {o.webname:<{max_len_out_name}} "
            f"{o.team:<{max_len_out_team}} {o.xP:<5.1f}"
            "  -->>  "
            f"{i.webname:<{max_len_in_name}} {i.team:<{max_len_in_team}} {i.xP:.1f}"
        )
    log.write(f"OxP gain: {(helpers.overall_xP(new) - helpers.overall_xP(old)):.1f}")
    log.write(f"TS  gain: {(helpers.tsscore(new) - helpers.tsscore(old)):.1f}")


def transfer(
    current: T.Sequence["structures.Player"],
    pool: T.Sequence["structures.Player"],
    remove: T.Sequence["structures.Player"],
    add: T.Sequence["structures.Player"],
    max_transfers: int,
    max_candidates: int,
    bar: tqdm,
):
    max_budget = helpers.squad_price(current)
    min_budget = max_budget * 0.8
    candidates = list[tuple[tuple[float, float, int], tuple[structures.Player, ...]]]()

    squad_base = {
        n: tuple(
            (c, helpers.squad_price(c))
            for c in sorted(
                itertools.combinations(current, len(current) - n),
                key=helpers.squad_price,
            )
        )
        for n in range(1, max_transfers + 1)
    }
    squad_base = {
        n: tuple(
            (players, cost)
            for players, cost in squad_cost
            if all(r not in players for r in remove)
        )
        for n, squad_cost in squad_base.items()
    }

    transfer_in = {
        n: tuple(
            (c, helpers.squad_price(c))
            for c in sorted(
                itertools.combinations(pool, n),
                key=helpers.squad_price,
            )
        )
        for n in range(1, max_transfers + 1)
    }
    transfer_in = {
        n: tuple(
            (players, cost)
            for players, cost in squad_cost
            if all(a in players for a in add)
        )
        for n, squad_cost in transfer_in.items()
    }

    min_max_tranfer_in = {
        n: (
            min(c for _, c in transfer_in[n]),
            max(c for _, c in transfer_in[n]),
        )
        for n in transfer_in
    }

    bar.total = sum(
        len(transfer_in[n]) * len(squad_base[n]) for n in range(1, max_transfers + 1)
    )

    sequence: int = 0

    for n in range(1, max_transfers + 1):
        for base, base_cost in squad_base[n]:
            min_c, max_c = min_max_tranfer_in[n]
            assert min_c <= max_c
            if max_c + base_cost < min_budget:
                bar.update(len(transfer_in[n]))
                continue
            if min_c + base_cost > max_budget:
                bar.update(len(transfer_in[n]))
                continue

            for t_in, t_in_cost in transfer_in[n]:
                cost = base_cost + t_in_cost
                if cost > max_budget:
                    break
                if cost < min_budget:
                    continue

                squad = base + t_in

                if add and all(a not in squad for a in add):
                    continue

                if (
                    helpers.valid_squad(squad)
                    and constraints.team_constraint(squad, 3)
                    and len(set(squad)) == 15
                ):
                    oxp = round(helpers.overall_xP(squad), 1)
                    inv = round(1 / (1 + helpers.tsscore(squad)), 3)
                    sequence += 1
                    if len(candidates) >= max_candidates:
                        heapq.heappushpop(
                            candidates,
                            (
                                (inv, oxp, sequence),
                                squad,
                            ),
                        )
                    else:
                        heapq.heappush(
                            candidates,
                            (
                                (inv, oxp, sequence),
                                squad,
                            ),
                        )

            bar.update(len(transfer_in[n]))

    return [
        structures.Squad(heapq.heappop(candidates)[-1]) for _ in range(len(candidates))
    ]


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="Transfer picker.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--add",
        nargs="+",
        default=[],
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--exclude",
        nargs="+",
        default=[],
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--max-candidates",
        default=100,
        type=int,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--max-transfers",
        type=int,
        required=True,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--min-mtm",
        default=0.0,
        type=float,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--min-xp",
        default=0.0,
        type=float,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--no-news",
        action="store_true",
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--remove",
        nargs="+",
        default=[],
        help="(default: %(default)s)",
    )

    args = parser.parse_args()

    pool = [p for p in fetch.players() if p.xP is not None]

    if args.exclude:
        pool = [p for p in pool if p.webname not in args.exclude]
        pool = [p for p in pool if p.team not in args.exclude]

    if args.min_xp > 0:
        pool = [p for p in pool if p.xP >= args.min_xp]

    if args.min_mtm > 0:
        pool = [p for p in pool if p.mtm() >= args.min_mtm]

    if args.no_news:
        pool = [p for p in pool if not p.news]

    pool = sorted(pool, key=lambda p: p.xP or 0)
    print(">>> Pool")
    print(structures.Squad(pool))

    print("\n>>>> Current team")
    team = fetch.my_team()
    print(team)

    with tqdm(
        bar_format="{percentage:3.0f}% | {bar:20} {r_bar}",
        unit_scale=True,
        unit_divisor=1_000,
        ascii=True,
    ) as bar:
        for new_squad in transfer(
            current=team.players,
            pool=pool,
            max_transfers=args.max_transfers,
            max_candidates=args.max_candidates,
            bar=bar,
            remove=[
                p
                for p in team.players
                if p.webname in args.remove or p.team in args.remove
            ]
            + [p for p in team.players if args.no_news and p.news],
            add=[p for p in pool if p.webname in args.add or p.team in args.add],
        ):
            display(team.players, new_squad, bar)


if __name__ == "__main__":
    main()



import collections
import typing as T

from lazyfpl import conf, structures


def squad_price(lineup: T.Sequence["structures.Player"]) -> int:
    return sum(p.price for p in lineup)


def squad_xP(lineup: T.Sequence["structures.Player"]) -> float:
    return sum(p.xP or 0 for p in lineup)


def overall_xP(lineup: T.Sequence["structures.Player"]) -> float:
    return (squad_xP(lineup) ** 2 + best_lineup_xP(lineup) ** 2) ** 0.5


def best_lineup(
    team: T.Sequence["structures.Player"],
    min_gkp: int = 1,
    min_def: int = 3,
    min_mid: int = 2,
    min_fwd: int = 1,
    size: int = 11,
) -> list["structures.Player"]:
    team = sorted(team, key=lambda x: x.xP or 0, reverse=True)
    gkps = [p for p in team if p.position == "GKP"]
    defs = [p for p in team if p.position == "DEF"]
    mids = [p for p in team if p.position == "MID"]
    fwds = [p for p in team if p.position == "FWD"]
    best = gkps[:min_gkp] + defs[:min_def] + mids[:min_mid] + fwds[:min_fwd]
    remainder = sorted(
        defs[min_def:] + mids[min_mid:] + fwds[min_fwd:],
        key=lambda x: x.xP or 0,
        reverse=True,
    )
    return best + remainder[: (size - len(best))]


def best_lineup_xP(lineup: T.Sequence["structures.Player"]) -> float:
    return squad_xP(best_lineup(lineup))


def valid_squad(
    squad: T.Sequence["structures.Player"],
    gkps: int = 2,
    defs: int = 5,
    mids: int = 5,
    fwds: int = 3,
) -> bool:
    cnt = collections.Counter(p.position for p in squad)
    return (
        cnt["GKP"] == gkps
        and cnt["DEF"] == defs
        and cnt["MID"] == mids
        and cnt["FWD"] == fwds
    )


def sscore(lineup: T.Sequence["structures.Player"], n: int = conf.lookahead) -> int:
    # "sscore -> "schedule score"
    # counts players in the lineup who plays in same match.
    # Ex. l'pool vs. man. city, and your team has Haaland and Salah as the only
    # players from the l'pool and city, the sscore is 2 since both play
    # the same match (assuming they start/play ofc.)

    per_gw = collections.defaultdict(list)
    for player in lineup:
        for i, nextopp in enumerate(player.upcoming_opponents()[:n]):
            per_gw[i].append((player.team, nextopp))

    return sum(sum(vs.count(x[::-1]) for x in set(vs)) for vs in per_gw.values())


def tcnt(lineup: T.Sequence["structures.Player"]) -> int:
    return sum(v - 1 for v in collections.Counter(p.team for p in lineup).values()) * 2


def tsscore(lineup: T.Sequence["structures.Player"], n: int = conf.lookahead) -> float:
    return (tcnt(lineup) ** 2 + sscore(lineup, n=n) ** 2) ** 0.5



import argparse
import dataclasses
import itertools
import math
import pickle
import traceback
import typing

import numpy as np
import torch
from torch.utils.data import DataLoader as TorchDataLoader
from torch.utils.data import Dataset as TorchDataset
from tqdm.std import tqdm

from lazyfpl import conf, database, fetch, populator, structures

if conf.debug:
    torch.set_printoptions(threshold=10_000)


@dataclasses.dataclass
class NormalizedFeatures:
    at_home: float
    points: float
    minutes: float
    opponent_strength_attack_away: float
    opponent_strength_attack_home: float
    opponent_strength_defence_away: float
    opponent_strength_defence_home: float
    opponent_strength_overall_away: float
    opponent_strength_overall_home: float
    team_strength_attack_away: float
    team_strength_attack_home: float
    team_strength_defence_away: float
    team_strength_defence_home: float
    team_strength_overall_away: float
    team_strength_overall_home: float


@dataclasses.dataclass
class FeatureBundle:
    features: tuple[NormalizedFeatures, ...]
    target: float


class Net(torch.nn.Module):
    def __init__(
        self,
        nfeature: int,
        rnn_hidden: int = 16,
    ) -> None:
        super().__init__()
        self.nfeature = nfeature
        self.rnn_hidden = rnn_hidden
        self.enc = torch.nn.GRU(
            input_size=nfeature,
            hidden_size=rnn_hidden,
            batch_first=True,
        )
        self.dec = torch.nn.Sequential(
            torch.nn.Linear(
                in_features=rnn_hidden,
                out_features=1,
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        _, h_final = self.enc(x)
        return self.dec(h_final.squeeze()).view(-1)


def features(f: "structures.Fixture") -> NormalizedFeatures:
    assert f.points is not None
    p_scale = database.points()
    s_scale = database.strengths()
    m_scale = database.minutes()
    return NormalizedFeatures(
        at_home=(f.at_home - 0.5) / 0.5,
        points=p_scale.normalize(f.points),
        minutes=m_scale.normalize(f.minutes or 0),
        opponent_strength_attack_away=s_scale.strength_attack_away.normalize(
            f.opponent_strength_attack_away,
        ),
        opponent_strength_attack_home=s_scale.strength_attack_home.normalize(
            f.opponent_strength_attack_home,
        ),
        opponent_strength_defence_away=s_scale.strength_defence_away.normalize(
            f.opponent_strength_defence_away,
        ),
        opponent_strength_defence_home=s_scale.strength_defence_home.normalize(
            f.opponent_strength_defence_home,
        ),
        opponent_strength_overall_away=s_scale.strength_overall_away.normalize(
            f.opponent_strength_overall_away,
        ),
        opponent_strength_overall_home=s_scale.strength_overall_home.normalize(
            f.opponent_strength_overall_home,
        ),
        team_strength_attack_away=s_scale.strength_attack_away.normalize(
            f.team_strength_attack_away,
        ),
        team_strength_attack_home=s_scale.strength_attack_home.normalize(
            f.team_strength_attack_home,
        ),
        team_strength_defence_away=s_scale.strength_defence_away.normalize(
            f.team_strength_defence_away,
        ),
        team_strength_defence_home=s_scale.strength_defence_home.normalize(
            f.team_strength_defence_home,
        ),
        team_strength_overall_away=s_scale.strength_overall_away.normalize(
            f.team_strength_overall_away,
        ),
        team_strength_overall_home=s_scale.strength_overall_home.normalize(
            f.team_strength_overall_home,
        ),
    )


class SequenceDataset(TorchDataset):
    def __init__(self, x: torch.Tensor, y: torch.Tensor) -> None:
        self.x, self.y = x, y

    def __len__(self) -> int:
        return self.x.shape[0]

    def __getitem__(self, idx: int) -> tuple[torch.Tensor, torch.Tensor]:
        return self.x[idx], self.y[idx]


def samples(
    fixtures: list["structures.Fixture"],
    upsample: int,
    backtrace: int = conf.backtrace,
) -> typing.Iterator[FeatureBundle]:
    fixtures = sorted(fixtures, key=lambda x: x.kickoff_time)
    # time --->
    train = [f for f in fixtures if not f.upcoming]
    min_ko = min(f.kickoff_time for f in train)
    max_ko = max(f.kickoff_time for f in train)

    if len(train) < backtrace:
        raise ValueError("To few samples.")

    while len(train) > backtrace:
        target = train.pop(-1)
        assert target.points is not None
        repat = max(
            (
                math.exp((target.kickoff_time - min_ko) / (max_ko - min_ko) * 2)
                / (math.e**2)
                * upsample
            ),
            1,
        )
        yield from itertools.repeat(
            FeatureBundle(
                features=tuple(features(f) for f in train[-backtrace:]),
                target=target.points,
            ),
            round(repat),
        )


def train(
    player: "structures.Player",
    epochs: int,
    lr: float,
    upsample: int,
    batch_size: int,
):
    bundles = tuple(samples(player.fixutres, upsample, conf.backtrace))
    ds = SequenceDataset(
        x=torch.tensor(
            tuple(tuple(dataclasses.astuple(f) for f in b.features) for b in bundles),
            dtype=torch.float32,
        ),
        y=torch.tensor(
            tuple(b.target for b in bundles),
            dtype=torch.float32,
        ),
    )
    loader = TorchDataLoader(ds, batch_size=batch_size, shuffle=True)

    loss_function = torch.nn.SmoothL1Loss()
    net = Net(ds[0][0].shape[-1])
    optimizer = torch.optim.AdamW(net.parameters(), lr=lr)

    for _ in range(epochs):
        for x, y in loader:
            optimizer.zero_grad()
            output = net(x)
            assert output.shape == y.shape, (output.shape, y.shape)
            loss = loss_function(output, y)
            loss.backward()
            optimizer.step()
    return net


def load_model(player: "structures.Player") -> "Net":
    pid = populator.player_id_fuzzer(player.name)
    if bts := database.load_model(pid):
        ms = pickle.loads(bts)
        n = Net(nfeature=ms["nfeature"], rnn_hidden=ms["rnn_hidden"])
        n.load_state_dict(ms["weights"])
        return n
    raise ValueError(f"No model for {player.name=} / {player.team=} / {pid=}.")


def save_model(player: "structures.Player", m: "Net") -> None:
    database.save_model(
        populator.player_id_fuzzer(player.name),
        pickle.dumps(
            {
                "rnn_hidden": m.rnn_hidden,
                "nfeature": m.nfeature,
                "weights": m.state_dict(),
            }
        ),
    )


def xP(
    player: "structures.Player",
    lookahead: int = conf.lookahead,
    backtrace: int = conf.backtrace,
) -> float:
    expected = list[float]()
    fixutres = sorted(player.fixutres, key=lambda x: x.kickoff_time)
    inference = [features(f) for f in fixutres if not f.upcoming][-backtrace:]
    upcoming = [f for f in fixutres if f.upcoming]

    model = load_model(player)
    model.eval()
    with torch.no_grad():
        for nxt in upcoming[:lookahead]:
            inf = torch.from_numpy(
                np.expand_dims(
                    np.stack(
                        [
                            np.array(dataclasses.astuple(x), dtype=np.float32)
                            for x in inference
                        ],
                        axis=0,
                    ).astype(np.float32),
                    axis=0,
                )
            )

            points = model(inf).detach().numpy()
            assert points.shape == (1,), points.shape
            expected.extend(points)
            inference.pop(0)
            inference.append(
                features(
                    structures.Fixture(
                        **(dataclasses.asdict(nxt) | {"points": points[0]})
                    )
                )
            )

    if conf.debug:
        print(player.name, player.team, expected)

    return round(sum(expected), 1)


def main():
    parser = argparse.ArgumentParser(
        prog="Player model trainer.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "--epochs",
        type=int,
        default=5,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--lr",
        type=float,
        default=0.01,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--min-mtm",
        type=int,
        default=0,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--upsample",
        type=int,
        default=25,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--batch_size",
        type=int,
        default=16,
        help="(default: %(default)s)",
    )

    args = parser.parse_args()

    players = [p for p in fetch.players() if p.mtm() >= args.min_mtm]

    with tqdm(
        ascii=True,
        bar_format="{percentage:3.0f}% | {bar:20} {r_bar}",
        total=len(players),
        unit_divisor=1_000,
        unit_scale=True,
    ) as bar:
        for player in players:
            try:
                m = train(
                    player,
                    epochs=args.epochs,
                    lr=args.lr,
                    upsample=args.upsample,
                    batch_size=args.batch_size,
                )
            except (
                IndexError,
                ValueError,
            ):
                ...
            except Exception as e:
                bar.write("".join(traceback.format_exception(e)))
            else:
                save_model(player, m)
                bar.write(
                    f"{xP(player):<6.1f} "
                    + f"{player.name} ("
                    + f"{player.team} - "
                    + f"{player.next_opponent})"
                )
            finally:
                bar.update(1)


if __name__ == "__main__":
    main()



# ruff: noqa: E501
import dataclasses
import datetime
import functools
import os
import pathlib
import sqlite3
import statistics
import typing as T

import pydantic

from lazyfpl import conf

SESSIONS = T.Literal["2021-22", "2022-23", "2023-24"]
CURRENT_SESSION = T.get_args(SESSIONS)[-1]
POSITIONS = T.Literal["GKP", "DEF", "MID", "FWD"]


@dataclasses.dataclass(frozen=True)
class SampleSummary:
    mean: float
    std: float
    variance: float

    @staticmethod
    def fromiter(values: T.Iterable[float]) -> "SampleSummary":
        return SampleSummary(
            mean=statistics.mean(values),
            std=statistics.stdev(values),
            variance=statistics.variance(values),
        )

    def normalize(self, value: float) -> float:
        return (value - self.mean) / self.variance


@dataclasses.dataclass(frozen=True)
class Strenghts:
    strength_attack_away: SampleSummary
    strength_attack_home: SampleSummary
    strength_defence_away: SampleSummary
    strength_defence_home: SampleSummary
    strength_overall_away: SampleSummary
    strength_overall_home: SampleSummary


class Game(pydantic.BaseModel):
    is_home: bool
    kickoff: datetime.datetime
    minutes: T.Optional[int]
    news: str
    opponent: str
    player_id: int
    player: str
    points: T.Optional[int]
    position: POSITIONS
    session: SESSIONS
    team: str
    upcoming: bool
    webname: str
    team_strength_attack_home: int
    team_strength_attack_away: int
    team_strength_defence_home: int
    team_strength_defence_away: int
    team_strength_overall_home: int
    team_strength_overall_away: int
    opponent_strength_attack_home: int
    opponent_strength_attack_away: int
    opponent_strength_defence_home: int
    opponent_strength_defence_away: int
    opponent_strength_overall_home: int
    opponent_trength_overall_away: int


@functools.cache
def connect(file: pathlib.Path = conf.db) -> sqlite3.Connection:
    return sqlite3.connect(file)


def execute(sql: str, parameters: tuple = ()) -> list[dict]:
    with connect() as conn:
        cursor = conn.execute(sql, parameters)
        if desc := [x[0] for x in cursor.description or []]:
            return [dict(zip(desc, row)) for row in cursor.fetchall()]
        return []


@functools.cache
def games() -> list[Game]:
    rows = execute(
        """
        select
            is_home,
            kickoff,
            minutes,
            points,
            position,
            session,
            upcoming,
            player_id,
            (select name from player where id = game.player_id)                    as player,
            (select news from player where id = game.player_id)                    as news,
            (select webname from player where id = game.player_id)                 as webname,
            (select name from team where opponent = team.id)                       as opponent,
            (select name from team where team = team.id)                           as team,
            (select strength_attack_home from team where team.id = game.team)      as team_strength_attack_home,
            (select strength_attack_away from team where team.id = game.team)      as team_strength_attack_away,
            (select strength_defence_home from team where team.id = game.team)     as team_strength_defence_home,
            (select strength_defence_away from team where team.id = game.team)     as team_strength_defence_away,
            (select strength_overall_home from team where team.id = game.team)     as team_strength_overall_home,
            (select strength_overall_away from team where team.id = game.team)     as team_strength_overall_away,
            (select strength_attack_home from team where team.id = game.opponent)  as opponent_strength_attack_home,
            (select strength_attack_away from team where team.id = game.opponent)  as opponent_strength_attack_away,
            (select strength_defence_home from team where team.id = game.opponent) as opponent_strength_defence_home,
            (select strength_defence_away from team where team.id = game.opponent) as opponent_strength_defence_away,
            (select strength_overall_home from team where team.id = game.opponent) as opponent_strength_overall_home,
            (select strength_overall_away from team where team.id = game.opponent) as opponent_trength_overall_away
        from
            game
    """
    )

    return [Game.model_validate(row) for row in rows]


def price(pid: int) -> int:
    rows = execute(
        """
        SELECT
            price
        FROM
            player
        WHERE
            id = ?
    """,
        (pid,),
    )
    return rows[0]["price"]


def webname(pid: int) -> str:
    rows = execute(
        """
        SELECT
            webname
        FROM
            player
        WHERE
            id = ?
    """,
        (pid,),
    )
    return rows[0]["webname"]


def save_model(player_id: int, model: bytes) -> None:
    execute(
        """
        UPDATE
            player
        SET
            model = ?
        WHERE
            id = ?
    """,
        (model, player_id),
    )


def load_model(player_id: int) -> bytes:
    return execute(
        """
        SELECT
            model
        FROM
            player
        WHERE
            id =?
    """,
        (player_id,),
    )[0]["model"]


@functools.cache
def points() -> SampleSummary:
    p = [
        row["points"]
        for row in execute(
            """
        SELECT
            points
        FROM
            game
        WHERE
            points is not null
    """
        )
    ]
    return SampleSummary.fromiter(p)


@functools.cache
def minutes() -> SampleSummary:
    p = [
        row["minutes"]
        for row in execute(
            """
        SELECT
            minutes
        FROM
            game
        WHERE
            minutes is not null
    """
        )
    ]
    return SampleSummary.fromiter(p)


@functools.cache
def strengths() -> Strenghts:
    rows = execute(
        """
        SELECT
            strength_attack_away,
            strength_attack_home,
            strength_defence_away,
            strength_defence_home,
            strength_overall_away,
            strength_overall_home
        FROM
            team
    """
    )
    return Strenghts(
        strength_attack_away=SampleSummary.fromiter(
            [r["strength_attack_away"] for r in rows]
        ),
        strength_attack_home=SampleSummary.fromiter(
            [r["strength_attack_home"] for r in rows]
        ),
        strength_defence_away=SampleSummary.fromiter(
            [r["strength_defence_away"] for r in rows]
        ),
        strength_defence_home=SampleSummary.fromiter(
            [r["strength_defence_home"] for r in rows]
        ),
        strength_overall_away=SampleSummary.fromiter(
            [r["strength_overall_away"] for r in rows]
        ),
        strength_overall_home=SampleSummary.fromiter(
            [r["strength_overall_home"] for r in rows]
        ),
    )



import argparse
import collections
import heapq
import itertools
import typing as T

from lazyfpl import constraints, fetch, structures, helpers
from tqdm.std import tqdm


def position_combinations(
    pool: list["structures.Player"],
    combinations: int,
) -> list[tuple[int, float, tuple["structures.Player", ...]]]:
    assert combinations > 0
    return sorted(
        (
            (helpers.squad_price(c), helpers.squad_xP(c), c)
            for c in itertools.combinations(pool, combinations)
        ),
        key=lambda x: (x[1], x[0]),
        reverse=True,
    )


def must_include(
    combinations: list[tuple[int, float, tuple["structures.Player", ...]]],
    include: list["structures.Player"],
) -> list[tuple[int, float, tuple["structures.Player", ...]]]:
    return (
        [c for c in combinations if all(i in c[-1] for i in include)]
        if include
        else combinations
    )


def lineup(
    pool: list["structures.Player"],
    budget_lower: int = 900,
    budget_upper: int = 1_000,
    include: T.Sequence["structures.Player"] = (),
    n_squads: int = 100_000,
    max_players_per_team: int = 3,
    alpha: float = 0.8,
) -> list[structures.Squad]:
    assert 0 <= alpha <= 1.0

    gkp_combinations = must_include(
        position_combinations(
            pool=[p for p in pool if p.position == "GKP"],
            combinations=2,
        ),
        [p for p in include if p.position == "GKP"],
    )

    def_combinations = must_include(
        position_combinations(
            pool=[p for p in pool if p.position == "DEF"],
            combinations=5,
        ),
        [p for p in include if p.position == "DEF"],
    )

    mid_combinations = must_include(
        position_combinations(
            pool=[p for p in pool if p.position == "MID"],
            combinations=5,
        ),
        [p for p in include if p.position == "MID"],
    )

    fwd_combinations = must_include(
        position_combinations(
            pool=[p for p in pool if p.position == "FWD"],
            combinations=3,
        ),
        [p for p in include if p.position == "FWD"],
    )

    # All combinations sorted from higest -> lowest xP.

    total = (
        len(gkp_combinations)
        * len(def_combinations)
        * len(mid_combinations)
        * len(fwd_combinations)
    )

    print(f"Goalkeeper combinations: {len(gkp_combinations):.1e}")
    print(f"Defender   combinations: {len(def_combinations):.1e}")
    print(f"Midfielder combinations: {len(mid_combinations):.1e}")
    print(f"Forwarder  combinations: {len(fwd_combinations):.1e}")
    print(f"Total      combinations: {total:.1e}")

    if not total:
        return []

    best_squads = list[tuple[tuple[float, float, int], tuple[structures.Player, ...]]]()
    best_squad_xp = (
        sum(
            (
                helpers.squad_xP(fwd_combinations[0][-1]),
                helpers.squad_xP(mid_combinations[0][-1]),
                helpers.squad_xP(def_combinations[0][-1]),
                helpers.squad_xP(gkp_combinations[0][-1]),
            )
        )
        * alpha
    )

    max_mid_price = max(price for price, _, _ in mid_combinations)
    min_mid_price = min(price for price, _, _ in mid_combinations)
    max_mid_xp = max(xp for _, xp, _ in mid_combinations)

    max_def_price = max(price for price, _, _ in def_combinations)
    min_def_price = min(price for price, _, _ in def_combinations)
    max_def_xp = max(xp for _, xp, _ in def_combinations)

    sequence: int = 0

    with tqdm(
        total=total,
        bar_format="{percentage:3.0f}% | {bar:20} {r_bar}",
        unit_scale=True,
        unit_divisor=1_000,
        ascii=True,
    ) as bar:
        for gp, gxp, g in gkp_combinations:
            for fp, fxp, f in fwd_combinations:
                bar.update(len(def_combinations) * len(mid_combinations))

                if gxp + fxp + max_def_xp + max_mid_xp < best_squad_xp:
                    # Could use break, but messes up tqdm.
                    continue

                if gp + fp + min_def_price + min_mid_price > budget_upper:
                    continue

                if gp + fp + max_def_price + max_mid_price < budget_lower:
                    continue

                if (
                    max(collections.Counter(p.team for p in g + f).values())
                    > max_players_per_team
                ):
                    continue

                for dp, dxp, d in def_combinations:
                    if gxp + fxp + dxp + max_mid_xp < best_squad_xp:
                        break

                    if gp + fp + dp + min_mid_price > budget_upper:
                        continue

                    if gp + fp + dp + max_mid_price < budget_lower:
                        continue

                    if (
                        max(collections.Counter(p.team for p in g + f + d).values())
                        > max_players_per_team
                    ):
                        continue

                    for mp, mxp, m in mid_combinations:
                        if gxp + fxp + dxp + mxp < best_squad_xp:
                            break

                        if (
                            budget_lower <= mp + dp + fp + gp <= budget_upper
                            and constraints.team_constraint(
                                squad := g + f + d + m, n=max_players_per_team
                            )
                            and (oxp := helpers.overall_xP(squad)) > best_squad_xp
                        ):
                            sequence += 1
                            if len(best_squads) >= n_squads:
                                heapq.heappushpop(
                                    best_squads,
                                    (
                                        (round(oxp), mp + dp + fp + gp, sequence),
                                        squad,
                                    ),
                                )
                                best_squad_xp = max(
                                    max(v for (_, v, _), _ in best_squads) * alpha,
                                    best_squad_xp,
                                )
                            else:
                                heapq.heappush(
                                    best_squads,
                                    (
                                        (round(oxp), mp + dp + fp + gp, sequence),
                                        squad,
                                    ),
                                )

    return [
        structures.Squad(heapq.heappop(best_squads)[-1])
        for _ in range(len(best_squads))
    ]


def position_price_candidates(
    pool: list["structures.Player"],
    topn: int,
) -> T.Iterator["structures.Player"]:
    for _, players in itertools.groupby(
        sorted(pool, key=lambda x: (x.position, x.price)),
        key=lambda x: (x.position, x.price),
    ):
        toadd = []
        team = set[str]()
        for player in sorted(players, key=lambda x: x.xP or 0, reverse=True):
            if player.team not in team:
                toadd.append(player)
                team.add(player.team)
        yield from toadd[:topn]


def main():
    parser = argparse.ArgumentParser(
        prog="Lineup optimizer",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "--budget-lower",
        type=int,
        default=900,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--budget-upper",
        type=int,
        default=1_000,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--gkp-def-not-same-team",
        action="store_true",
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--include",
        nargs="+",
        default=[],
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--keep-squad",
        type=int,
        default=100_000,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--max-def-per-team",
        type=int,
        default=3,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--max-players-per-team",
        type=int,
        default=3,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--min-mtm",
        type=float,
        default=0.0,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--min-xp",
        type=float,
        default=0.0,
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--no-news",
        action="store_true",
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--remove",
        nargs="+",
        default=[],
        help="(default: %(default)s)",
    )
    parser.add_argument(
        "--top-position-price",
        type=int,
        default=0,
        help="(default: %(default)s)",
    )

    args = parser.parse_args()

    pool = [p for p in fetch.players() if p.xP is not None]
    pool = [p for p in pool if p.mtm() >= args.min_mtm and p.xP >= args.min_xp]

    if args.no_news:
        pool = [p for p in pool if not p.news]

    player_names = {p.webname.lower() for p in pool}
    remove: set[str] = {r.lower() for r in args.remove if r.lower() in player_names}
    pool = [p for p in pool if p.webname.lower() not in remove]

    team_names = {p.team.lower() for p in pool}
    remove_team: set[str] = {r.lower() for r in args.remove if r.lower() in team_names}
    pool = [p for p in pool if p.team.lower() not in remove_team]

    if args.top_position_price:
        pool = list(
            set(
                position_price_candidates(
                    pool=pool,
                    topn=args.top_position_price,
                )
            )
        )

    include = list(
        {
            p
            for p in fetch.players()
            if p.webname in args.include
            or p.name in args.include
            or p.team in args.include
        }
    )
    pool += include

    # Just in case
    pool = list(set(pool))

    print(structures.Squad(pool))
    squads = lineup(
        pool=pool,
        budget_lower=args.budget_lower,
        budget_upper=args.budget_upper,
        include=include,
        max_players_per_team=args.max_players_per_team,
        n_squads=args.keep_squad,
    )

    if args.gkp_def_not_same_team:
        squads = [s for s in squads if constraints.gkp_def_same_team(s.players)]

    if args.max_def_per_team:
        squads = [
            s
            for s in squads
            if max(
                collections.Counter(p.team for p in s if p.position == "DEF").values()
            )
            <= args.max_def_per_team
        ]

    for player, cnt in collections.Counter(
        p for squad in squads for p in squad.players
    ).items():
        print(player, round(cnt / len(squads) * 100, 2))

    print("\n\n".join(str(s) for s in squads[-10:]))

    mincxp = min(s.CxP() for s in squads)
    maxxcp = max(s.CxP() for s in squads)
    minss = min(s.sscore() for s in squads)
    maxss = max(s.sscore() for s in squads)
    mints = min(s.tsscore() for s in squads)
    maxts = max(s.tsscore() for s in squads)

    print("")
    print(
        f"Min CxP: {mincxp:.1f}",
        f"Max CxP: {maxxcp:.1f}",
        f"Max-Min CxP: {(maxxcp-mincxp):.1f}",
    )
    print(
        f"Min ss: {minss:d}",
        f"Max ss: {maxss:d}",
        f"Max-Min ss: {(maxss-minss):d}",
    )
    print(
        f"Min ts: {mints:.2f}",
        f"Max ts: {maxts:.2f}",
        f"Max-Min ts: {(maxts-mints):.2f}",
    )


if __name__ == "__main__":
    main()
